#include <stdio.h>
#include <stdlib.h>
#include "matrix.h"
#include "lgs.h"


/* externes Klassenobjekt einbinden */

extern matrix nm_matrix;


/* Hauptprogramm */

int main(int argc, char *argv[])
{
  FILE *in = stdin;     /* Ein-/Ausgabe */
  FILE *out = stdout;
  FILE *temp;
  int retcode;     
  int hilfe = 0;        /* 1 = Hilfe fuer Benutzer anzeigen */   
  int fehler = 0;       /* 1 = Programm falsch aufgerufen */
  int schalter = 1;     /* 0 = es folgen keine Schalter ('--' gefunden) */

  linearesSystem lgs;   /* nimmt Parameter und Ergebnis auf */ 
  matrix *lmalr;

  while ((schalter) && (argc>1) && (*argv[1]=='-')) {/* Schalter bearbeiten */
    switch ((argv[1])[1]) {
      case 'h': hilfe = 1;
                break;
      case '-': schalter = 0;
                break;
      default:  fprintf(stderr,"Schalter %s unbekannt.\n", argv[1]);
                fehler = 1;
                break;
    }
    argv[1] = argv[0];
    argv++; 
    argc--;
  }
  if ((argc>3) || (argc<2)) fehler = 1;
  if ((hilfe) || (fehler)) {             /* Hilfetext anzeigen */
    fprintf(stderr,"\nBenutzung: "
                   "%s [Schalter] [--] Eingabedatei [Ausgabedatei]\n\n",
                   argv[0]);
    fprintf(stderr,"Schalter: derzeit noch keine ausser -h fuer diese Hilfe\n");
    fprintf(stderr,"Zum Format der Ein-/Ausgabe siehe Beispieldateien\n");
    fprintf(stderr,"\nMirko Belickin und Joachim Wagner - 28. Oktober 1999\n");
    exit(fehler);
  }
  if (argc>1) {                         
    if (!(temp=fopen(argv[1],"rb"))) {
      fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[1]);
      exit(1);
    }
    in = temp;
  }
  if (argc>2) {
    if (!(temp=fopen(argv[2],"wb"))) {
      fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[2]);
      exit(1);
    }
    out = temp;
  }
  lgs.a = nm_matrix.aktion->laden(in, NULL);    /* A und b aus Datei */
  lgs.b = nm_matrix.aktion->laden(in, NULL);    /* einlesen          */

  if (!lrZerlegungUndDeterminanteUndLoesen(     /* Zerlegung durchfuehren */
          &lgs, 
          relativeSpaltenMaximumStrategie)) {   /* Erfolg? */

    fprintf(stderr,"Matrix konnte nicht zerlegt werden\n");  /* Nein ... */

  }else{                                        /* Ja -> Ergebnis ausgeben */

    fprintf(stdout,"\nA nach Zerlegung (L und R kombiniert in einer Matrix):\n");
    lgs.a->aktion->speichern(lgs.a, stdout);    /* veraendertes A ausgeben */

    fprintf(stdout,"\nMatrix L:\n");
    lgs.l->aktion->speichern(lgs.l, stdout);    /* L ausgeben */

    fprintf(stdout,"\nMatrix R:\n");
    lgs.r->aktion->speichern(lgs.r, stdout);    /* R ausgeben */

    fprintf(stdout,"\nL * R: (sollte in etwa eine Zeilenpermutation von A ergeben.)\n");
    lmalr = nm_matrix.aktion->produkt(lgs.l, lgs.r);
    lmalr->aktion->speichern(lmalr, stdout);    /* L mal R ausgeben */

    fprintf(stdout,"\nLoesungsvektor x:\n");
    lgs.x->aktion->speichern(lgs.x, stdout);    /* x ausgeben */

    fprintf(stdout,"\ndet(A) = %f\n", lgs.det); /* Determinate ausgeben */
  }

  if ( ferror(in) ) {
    fprintf(stderr, "Fehler in der Eingabe.\n");
    retcode = 1;
  }else{
    retcode = 0;
  }  
  if (in  != stdin)  fclose(in);                /* Dateien ggf. schliessen */
  if (out != stdout) fclose(out);

  return retcode;
}
