/* lgs.c
 * loest ein lineares Gleichungssystem und gibt zusaetzliche Infos
 * erstellt: Mit Okt 27 15:09:30 CEST 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include "matrix.h"
#include "nmmatrix.h"
#include "lgs.h"
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/* Einbinden des externen Klassenobjekts NM_MATRIX */

extern matrix nm_matrix;


/* Deklaration der Hilfsfunktionen */

static int lrZerlegungUndDeterminante(linearesSystem *lgs,
                                      pivotStrategieFunktion pivotStrategie);

static int vorwaertssubstitution(linearesSystem *lgs);

static int rueckwaertssubstitution(linearesSystem *lgs);


/* ------------------------------------------------------------------------ 
 * Implementation   
 */ 


int lrZerlegungUndDeterminanteUndLoesen(linearesSystem* lgs, 
                                        pivotStrategieFunktion pivotStrategie)
{
  int i, j;
  nmmatrixschablone schablone;

  /* A und b pruefen */

  if ((!lgs->a) || (!lgs->b))
    return 0;
  if (lgs->a->aktion->breite(lgs->a) != lgs->a->aktion->hoehe(lgs->a))
    return 0;
  if (lgs->b->aktion->breite(lgs->b) != 1)   
    return 0;
  if (lgs->a->aktion->breite(lgs->a) != lgs->b->aktion->hoehe(lgs->b))
    return 0;
      
  /* Schritte des Gaussverfahrens: */

  if (!lrZerlegungUndDeterminante(lgs, pivotStrategie))
    return 0;                     /* Fehler: Matrix singulaer */
  vorwaertssubstitution(lgs);
  rueckwaertssubstitution(lgs);

 
  /* Kompakt gespeicherte LR-Zerlegung in Matrizen l und r kopieren */
    
  schablone.n = lgs->a->aktion->hoehe(lgs->a);
  schablone.m = lgs->a->aktion->breite(lgs->a);
  schablone.init = 0;
  lgs->l = nm_matrix.aktion->kreieren(&schablone, NULL);
  lgs->r = nm_matrix.aktion->kreieren(&schablone, NULL);
  
  for (i=1; i<=schablone.n; i++) {
    for (j=1; j<=schablone.m; j++) {
      if (i>j) {                                 /* unterhalb der Diagonale */
        lgs->l->aktion->setzen(lgs->l, i, j,     /* l(i,j) = a(i,j) */
          lgs->a->aktion->amIndex(lgs->a, i, j));
        lgs->r->aktion->setzen(lgs->r, i, j, 0); /* r(i,j) = 0; */
      }
      if (i==j) {                                /* in der Diagonale */
        lgs->l->aktion->setzen(lgs->l, i, j, 1); /* l(i,j) = 1 */
        lgs->r->aktion->setzen(lgs->r, i, j,     /* r(i,j) = a(i,j); */
          lgs->a->aktion->amIndex(lgs->a, i, j));
      }
      if (i<j) {                                 /* oberhalb der Diagonale */
        lgs->l->aktion->setzen(lgs->l, i, j, 0); /* l(i,j) = 0 */
        lgs->r->aktion->setzen(lgs->r, i, j,     /* r(i,j) = a(i,j); */
          lgs->a->aktion->amIndex(lgs->a, i, j));
      }
    }
  }
  return 1; /* Erfolg */
}


static int lrZerlegungUndDeterminante(linearesSystem *lgs,
                                      pivotStrategieFunktion pivotStrategie)
{
  int i, j, k, n, pk;
  double h;                       /* Hilfsvariable zum Tauschen */
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.init = 1;
  lgs->p = nm_matrix.aktion->kreieren(&schablone, NULL);
  lgs->det = 1.0;
  for (k=1; k<=(n-1); k++) {
    pk = pivotStrategie(lgs->a, k);  /* Pivotzeile bestimmen */
    if (!pk) {                       /* Matrix singulaer? */
      return 0;                      /* Fehler anzeigen */
    }
    lgs->p->aktion->setzen(lgs->p,k,1,pk);     /* und in Vektor p eintragen */
    if (pk != k) {                   /* Sollen Zeilen getauscht werden? */ 
      lgs->det *= -1;                /* Vorzeichen der Determinate wechseln */
      for (j=1; j<=n; j++) {         /* Tauschen */
        h = lgs->a->aktion->amIndex(lgs->a, k, j);
        lgs->a->aktion->setzen(lgs->a, k, j, 
          lgs->a->aktion->amIndex(lgs->a, pk, j));
        lgs->a->aktion->setzen(lgs->a, pk, j, h);
      }
    }
    lgs->det *=                      /* Determinante fortsetzen*/ 
      lgs->a->aktion->amIndex(lgs->a, k, k); 
    for (i=k+1; i<=n; i++) {
      lgs->a->aktion->setzen(lgs->a, i, k,     /* l(i,k) bestimmen */
          lgs->a->aktion->amIndex(lgs->a, i, k)
          / lgs->a->aktion->amIndex(lgs->a, k, k));
      for (j=k+1; j<=n; j++) {
        lgs->a->aktion->setzen(lgs->a, i, j,   /* vordersten Koeffizienten */
            lgs->a->aktion->amIndex(lgs->a, i, j) /* auf Null bringen */
            - lgs->a->aktion->amIndex(lgs->a, i, k)
              * lgs->a->aktion->amIndex(lgs->a, k, j));
      }
    }
  }
  lgs->det *=                        /* Determinatenberechnung abschliessen */
    lgs->a->aktion->amIndex(lgs->a, n, n); 
  return 1;                          /* Erfolg melden */
}


/* Wir stellen dem Benutzer folgende Strategien bereit: */

int diagonalStrategie(matrix *m, int k) 
{
  if (fabs(m->aktion->amIndex(m,k,k)) < pow(2,-120)) 
    return 0;
  else
    return k;
}


int relativeSpaltenMaximumStrategie(matrix *m, int k)
{
  int i, j, n, pk;
  double max, q, s;
  n = m->aktion->breite(m); 
  max = 0.0;
  for (i=k; i<=n; i++) {      /* ab Zeile k nach Pivotzeile suchen */
    s = 0.0;
    for (j=k; j<=n; j++)      /* Zeilensumme berechnen */
      s += fabs(m->aktion->amIndex(m,i,j));
    q = fabs(m->aktion->amIndex(m,i,k)) /s;  /* relativieren */
    if (q > max) {            /* neue, besser geeignete Zeile gefunden */
      max = q;
      pk = i;       
    }
  }
  if (max < pow(2,-120))      /* Wenn max kleiner epsilon, das hier noch */
    return 0;                 /* deutlich ueber der kleinsten darstellbaren */
                              /* Zahl liegt, dann STOP */
  else
    return pk;                /* Pivotzeile zureuckgeben */
} 


/* 2. Schritt des Gauss-Verfahrens */

static int vorwaertssubstitution(linearesSystem *lgs)
{
  int i, j, k, n, pk;
  double h;
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.init = 1;
  lgs->c = nm_matrix.aktion->kreieren(&schablone, NULL);  /* neuer Vektor c */
  for (k=1; k<=(n-1); k++) {
    pk = lgs->p->aktion->amIndex(lgs->p, k, 1);
    if (pk != k) {
      h = lgs->b->aktion->amIndex(lgs->b, k, 1);   /* Tausch */
      lgs->b->aktion->setzen(lgs->b, k, 1, 
        lgs->b->aktion->amIndex(lgs->b, pk, 1));
      lgs->b->aktion->setzen(lgs->b, pk, 1, h);
    }
  }
  for (i=1; i<=n; i++) {
    lgs->c->aktion->setzen(lgs->c, i, 1,           /* c(i) = b(i)           */
      lgs->b->aktion->amIndex(lgs->b, i, 1));
    for (j=1; j<=(i-1); j++) {
      lgs->c->aktion->setzen(lgs->c, i, 1,         /* c(i) =                */
        lgs->c->aktion->amIndex(lgs->c, i, 1)      /*        c(i)           */
        - lgs->a->aktion->amIndex(lgs->a, i, j)    /*        - a(i,j)       */
          * lgs->c->aktion->amIndex(lgs->c, j, 1));/*          * c(j)       */
    }
  }
  return 1;   /* Erfolg */
}


/* 3. Schritt des Gauss-Verfahrens */

static int rueckwaertssubstitution(linearesSystem *lgs)
{
  int i, k, n;
  double s;
  nmmatrixschablone schablone;
  n = lgs->a->aktion->breite(lgs->a); 
  schablone.n = n;
  schablone.m = 1;
  schablone.init = 1;
  lgs->x = nm_matrix.aktion->kreieren(&schablone, NULL);
  for (i=n; i>=1; i--) {                       /* i von Zeile n bis 1  */
    s = lgs->c->aktion->amIndex(lgs->c, i, 1);       /* c(i) +         */
    for (k=(i+1); k<=n; k++) {                       /* Summe          */
      s += lgs->a->aktion->amIndex(lgs->a, i, k)     /*   a(i,k)       */
           * lgs->x->aktion->amIndex(lgs->x, k, 1);  /*   * x(k) von   */
    }                                                /* k=i+1 bis n    */
    s /= lgs->a->aktion->amIndex(lgs->a, i, i); /* Zeile normalisieren */
    lgs->x->aktion->setzen(lgs->x, i, 1, -s);   /* und x(i) ist freigestellt */
  }
  return 0;
}
