/* nmmatrix.c
 * Matrixtyp nxm-Matrix
 * erstellt: Die Okt 26 16:50:22 CEST 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include "matrix.h"
#include "namen.h"
#include "nmmatrix.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/* Deklaration der Objekt-Methoden */

matrix * nmmatrix_laden(FILE *datei, void *vater);
int nmmatrix_speichern(matrix *m, FILE *datei);
matrix * nmmatrix_kreieren(void *schablone, void *vater);
int nmmatrix_freigeben(matrix *m, void *vater);
double nmmatrix_amIndex(matrix *m, int zeile, int spalte);
void nmmatrix_setzen(matrix *m, int zeile, int spalte, double d);
int nmmatrix_breite(matrix *m);
int nmmatrix_hoehe(matrix *m);
matrix * nmmatrix_produkt(matrix *a, matrix *b);


/* Globales Klassenobjekt */

matrixFunktionen nm_matrixFunktionen = {
    nmmatrix_laden,
    nmmatrix_speichern,
    nmmatrix_freigeben,
    nmmatrix_kreieren,
    nmmatrix_amIndex,
    nmmatrix_setzen,
    nmmatrix_breite,
    nmmatrix_hoehe,
    nmmatrix_produkt
};

matrix nm_matrix = {          /* das Klassenobjekt zur Klasse NM_MATRIX */
    &nm_matrixFunktionen,     
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* private Methoden */

static nmmatrixdaten *get_nmmatrixdaten(matrix *m);
static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum);


/* ------------------------------------------------------------------------ 
 * nxm-Matrix Implementation
 */ 


matrix * nmmatrix_laden(FILE *datei, void *vater)
{
  char eingabe[BUFSIZ];
  nmmatrixschablone schablone;
  int i;
  matrix *m;
  nmmatrixdaten *pm;
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Keine Daten fuer neue Matrix gefunden - EOF.\n");
    exit(1);
  }
  if (strncmp(eingabe,"NM_MATRIX:",
               strlen("NM_MATRIX:"))!=0) {
    fprintf(stderr,"Dies ist keine NM_Matrix!\n");
    exit(1);
  }
  /* spaeter: hier Name der Matrix extrahieren */

  /* Abmasse der Matrix feststellen */

  lese_datum(datei, "%d", "n =", &(schablone.n));
  lese_datum(datei, "%d", "m =", &(schablone.m));

  /* leere Matrix erzeugen */

  schablone.init = 0;
  m = nm_matrix.aktion->kreieren(&schablone, vater);

  /* Eintraege aus der Datei lesen */

  pm = get_nmmatrixdaten(m);

  for (i = 0; i < pm->n * pm->m; i++)
    lese_datum(datei, "%lf", "", &(pm->eintrag[i]));

  return m;
}

static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum)
{
  char eingabe[BUFSIZ];
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Zeile %s nicht gefunden - EOF.\n", prefix);
    exit(1);
  }
  if (strncmp(eingabe, prefix,
               strlen( prefix))!=0) {
    fprintf(stderr,"%s... erwartet!\n", prefix);
    exit(1);
  }
  sscanf(&(eingabe[strlen(prefix)]), format, pdatum);
                               /* erstmal ohne Fehlerkontrolle */
  return 0;
}

int nmmatrix_speichern(matrix *m, FILE *datei)
{
  nmmatrixdaten *pm;
  int i;
  pm = get_nmmatrixdaten(m);
  fprintf(datei,"NM_MATRIX: %s\n", "");/* optionaler Name (spaeter) */
  fprintf(datei,"n = %d\n", pm->n);
  fprintf(datei,"m = %d\n", pm->m);
  for (i = 0; i < pm->n * pm->m; i++)
    fprintf(datei,"%f\n", pm->eintrag[i]);
  return 0;
}

matrix * nmmatrix_kreieren(void *schablone, void *vater)
{
  nmmatrixschablone *s = (nmmatrixschablone *) schablone;
  int i;
  matrix *m;
  nmmatrixdaten *pm;
  if (!(m = (matrix *)malloc(sizeof(matrix)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue Matrix (Teil 1)!\n");
    exit(1);          
  }
  m->aktion = nm_matrix.aktion;    /* moegliche Aktionen festlegen */
  if (!(m->daten = (nmmatrixdaten *)malloc(sizeof(nmmatrixdaten)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue Matrix (Teil 2)!\n");
    exit(1);          
  }
  pm = get_nmmatrixdaten(m);

  pm->n = s?s->n:1; 
  pm->m = s?s->m:1; 
  if (!(pm->eintrag = (double *)malloc( sizeof(double)
                                        * pm->n * pm->m ))) {
    fprintf(stderr,"Nicht genug Speicher fuer Matrixeintraege!\n");
    exit(1);          
  }
  if (s && s->init) {
    for(i=0; i < pm->n * pm->m; i++)
      pm->eintrag[i] = 0.0;
  }
  return m;
}

int nmmatrix_freigeben(matrix *m, void *vater) 
{ 
  nmmatrixdaten *pm;
  pm = get_nmmatrixdaten(m);
  free(pm->eintrag);     
  free(pm);
  free(m);
  return 0;              /* immer Erfolg */
}

static nmmatrixdaten *get_nmmatrixdaten(matrix *m) 
{
  if ((!(m->daten))) /* sicher ist sicher ... */
    exit(1);                        /* Fehlermeldung fehlt noch */
  return m->daten;                  /* Daten der Matrix zurueckgeben */
}


#define nmmatrix_pruefeIndex(pm, zeile, spalte)  \
  if ((zeile < 1) ||  \
      (spalte < 1) || \
      (zeile > pm->n) || \
      (spalte > pm->m)) { \
   fprintf(stderr,"Ungueltiger Index fuer Matrix: (%i,%i)\n", \
                  zeile, spalte); \
   exit(1); \
  } 


double nmmatrix_amIndex(matrix *m, int zeile, int spalte)
{
  nmmatrixdaten *pm;
  pm = get_nmmatrixdaten(m);
  nmmatrix_pruefeIndex(pm, zeile, spalte); 
  zeile--, spalte--;
  return pm->eintrag[ pm->m * zeile + spalte ];    /* erstmal ohne Kontrolle */
}

void nmmatrix_setzen(matrix *m, int zeile, int spalte, double d)
{
  nmmatrixdaten *pm;
  pm = get_nmmatrixdaten(m);
  nmmatrix_pruefeIndex(pm, zeile, spalte); 
  zeile--, spalte--;
  pm->eintrag[ pm->m * zeile + spalte ] = d;       /* erstmal ohne Kontrolle */
}

int nmmatrix_breite(matrix *m)
{
  nmmatrixdaten *pm;
  pm = get_nmmatrixdaten(m);
  return pm->m;
}

int nmmatrix_hoehe(matrix *m)
{
  nmmatrixdaten *pm;
  pm = get_nmmatrixdaten(m);
  return pm->n;
}

matrix * nmmatrix_produkt(matrix *m1, matrix *m2)
{
  nmmatrixschablone schablone;
  int i, j, k, m1breite;
  matrix *m;
  nmmatrixdaten *pm;
  double s;
  m1breite = m1->aktion->breite(m1);
  if (m1breite != m2->aktion->hoehe(m2)) {
    fprintf(stderr,"Matrizen m1 und m2 koennen nicht multipliziert werden.\n");
    exit(1);
  }
  schablone.n = m1->aktion->hoehe(m1);
  schablone.m = m2->aktion->breite(m2);
  schablone.init = 0;
  m = nm_matrix.aktion->kreieren(&schablone, NULL); /* neue Produktmatrix */
  pm = get_nmmatrixdaten(m);
  for (i=1; i<=pm->n; i++) {             /* Produkt berechnen */
    for (j=1;  j<=pm->m; j++) {
      s = 0.0;
      for (k=1; k<=m1breite; k++)        /* Summe m1(i,k) * m2(k,j) ueber k */
        s += m1->aktion->amIndex(m1,i,k)
             * m2->aktion->amIndex(m2,k,j);
      pm->eintrag[ pm->m * (i-1) + (j-1) ] = s;  /* in m(i,j) ablegen */
    }
  }
  return m;                          
}
